@@ -9,15 +9,17 @@ written by Henning Jacobs <henning@jacobs1.de>
 
 from __future__ import print_function
 
-from StringIO import StringIO
+try:
+    from StringIO import StringIO
+except ImportError:
+    # Python 3
+    from io import StringIO
 from collections import defaultdict
 
-from pythontidy import PythonTidy
 from tempfile import NamedTemporaryFile
 from xml.etree.ElementTree import ElementTree
 from xml.etree.ElementTree import fromstring as xmlfromstring
 import argparse
-import ast
 import contextlib
 import csv
 import fnmatch
@@ -30,11 +32,20 @@ import sys
 import tempfile
 import shutil
 
+if sys.version_info.major == 2:
+    # Pythontidy is only supported on Python2
+    from pythontidy import PythonTidy
+
+running_on_py3 = sys.version_info.major == 3
+
+
 NOT_SPACE = re.compile('[^ ]')
 
 TRAILING_WHITESPACE_CHARS = set(' \t')
 INDENTATION = '    '
 
+DEFAULT_CONFIG_PATHS = ['~/.codevalidatorrc', '/etc/codevalidatorrc']
+
 DEFAULT_RULES = [
     'utf8',
     'nobom',
@@ -47,10 +58,14 @@ DEFAULT_CONFIG = {
     'exclude_dirs': ['.svn', '.git'],
     'exclude_files': ['.*.swp'],
     'rules': {
+        '*.c': DEFAULT_RULES,
         '*.coffee': DEFAULT_RULES + ['coffeelint'],
         '*.conf': DEFAULT_RULES,
+        '*.cpp': DEFAULT_RULES,
         '*.css': DEFAULT_RULES,
+        '*.erb': DEFAULT_RULES + ['erb'],
         '*.groovy': DEFAULT_RULES,
+        '*.h': DEFAULT_RULES,
         '*.htm': DEFAULT_RULES,
         '*.html': DEFAULT_RULES,
         '*.java': DEFAULT_RULES + ['jalopy'],
@@ -63,11 +78,13 @@ DEFAULT_CONFIG = {
         '*.phtml': DEFAULT_RULES,
         '*.pp': DEFAULT_RULES + ['puppet'],
         '*.properties': DEFAULT_RULES + ['ascii'],
-        '*.py': DEFAULT_RULES + ['pep8', 'pyflakes'],
+        '*.py': DEFAULT_RULES + ['pyflakes', 'pythontidy'],
         '*.rst': DEFAULT_RULES,
+        '*.rb': DEFAULT_RULES + ['ruby', 'rubocop'],
+        '* *': ['invalidpath'],
         '*.sh': DEFAULT_RULES,
-        '*.sql': DEFAULT_RULES + ['sql_comment_last_line', 'sql_semi_colon'],
-        '*.sql_diff': DEFAULT_RULES,
+        '*.sql': DEFAULT_RULES + ['sql_semi_colon'],
+        '*.sql_diff': DEFAULT_RULES + ['sql_semi_colon'],
         '*.styl': DEFAULT_RULES,
         '*.txt': DEFAULT_RULES,
         '*.vm': DEFAULT_RULES,
@@ -77,12 +94,8 @@ DEFAULT_CONFIG = {
         '*.yml': DEFAULT_RULES + ['yaml'],
         '*pom.xml': ['pomdesc'],
     },
-    'options': {'phpcs': {'standard': 'PSR', 'encoding': 'UTF-8'}, 'pep8': {
-        'max_line_length': 120,
-        'ignore': 'N806',
-        'passes': 5,
-        'select': 'e501',
-    }, 'jalopy': {'classpath': '/opt/jalopy/lib/jalopy-1.9.4.jar:/opt/jalopy/lib/jh.jar'}},
+    'options': {'phpcs': {'standard': 'PSR', 'encoding': 'UTF-8'}, 'pep8': {'max_line_length': 120, 'ignore': 'N806',
+                'passes': 5}, 'jalopy': {'classpath': '/opt/jalopy/lib/jalopy-1.9.4.jar:/opt/jalopy/lib/jh.jar'}},
     'dir_rules': {'db_diffs': ['sql_diff_dir', 'sql_diff_sql'], 'database': ['database_dir']},
     'create_backup': True,
     'backup_filename': '.{original}.pre-cvfix',
@@ -152,6 +165,14 @@ def message(msg):
     return wrap
 
 
+def is_python3(fd):
+    '''check first line of file object whether it contains "python3" (shebang)'''
+
+    line = fd.readline()
+    fd.seek(0)
+    return b'python3' in line
+
+
 @message('has invalid file path (file name or extension is not allowed)')
 def _validate_invalidpath(fd):
     return False
@@ -159,20 +180,24 @@ def _validate_invalidpath(fd):
 
 @message('contains tabs')
 def _validate_notabs(fd):
-    return '\t' not in fd.read()
+    return b'\t' not in fd.read()
 
 
 def _fix_notabs(src, dst):
-    dst.write(src.read().replace('\t', ' ' * 4))
+    original = src.read()
+    fixed = original.replace(b'\t', b' ' * 4)
+    dst.write(fixed.decode())
 
 
 @message('contains carriage return (CR)')
 def _validate_nocr(fd):
-    return '\r' not in fd.read()
+    return b'\r' not in fd.read()
 
 
 def _fix_nocr(src, dst):
-    dst.write(src.read().replace('\r', ''))
+    original = src.read()
+    fixed = original.replace(b'\r', b'')
+    dst.write(fixed.decode())
 
 
 @message('is not UTF-8 encoded')
@@ -195,7 +220,7 @@ def _validate_ascii(fd):
 
 @message('has UTF-8 byte order mark (BOM)')
 def _validate_nobom(fd):
-    return not fd.read(3).startswith('\xef\xbb\xbf')
+    return not fd.read(3).startswith(b'\xef\xbb\xbf')
 
 
 @message('contains invalid indentation (not 4 spaces)')
@@ -214,6 +239,7 @@ def _validate_indent4(fd):
 @message('contains lines with trailing whitespace')
 def _validate_notrailingws(fd):
     for line in fd:
+        line = line.decode() if running_on_py3 else line
         if line.rstrip('\n\r')[-1:] in TRAILING_WHITESPACE_CHARS:
             return False
     return True
@@ -238,7 +264,7 @@ def _validate_xml(fd):
     tree = ElementTree()
     try:
         tree.parse(fd)
-    except Exception, e:
+    except Exception as e:
         _detail('%s: %s' % (e.__class__.__name__, e))
         return False
     return True
@@ -257,7 +283,7 @@ def _fix_xmlfmt(src, dst):
 def _validate_json(fd):
     try:
         json.load(fd)
-    except Exception, e:
+    except Exception as e:
         _detail('%s: %s' % (e.__class__.__name__, e))
         return False
     return True
@@ -267,8 +293,13 @@ def _validate_json(fd):
 def _validate_yaml(fd):
     import yaml
     try:
-        yaml.safe_load(fd)
-    except Exception, e:
+        # Using safeloader because it supports recursive nodes
+        loader = yaml.SafeLoader(fd)
+        # Support random tags
+        loader.add_multi_constructor('!', (lambda _, tag, _2: tag)) 
+        while loader.check_data():
+            loader.get_data()
+    except Exception as e:
         _detail('%s: %s' % (e.__class__.__name__, e))
         return False
     return True
@@ -276,6 +307,9 @@ def _validate_yaml(fd):
 
 @message('is not PythonTidy formatted')
 def _validate_pythontidy(fd):
+    if is_python3(fd) or running_on_py3:
+        # PythonTidy supports Python 2 only
+        return True
     source = StringIO(fd.read())
     if len(source.getvalue()) < 4:
         # small or empty files are ignored
@@ -286,39 +320,63 @@ def _validate_pythontidy(fd):
 
 
 @message('is not pep8 formatted')
-def _validate_pep8(fd, options):
+def _validate_pep8(fd, options={}):
     import pep8
-    pep8style = pep8.StyleGuide(max_line_length=options['max_line_length'])
+
+    # if user doesn't define a new value use the pep8 default
+    max_line_length = options.get('max_line_length', pep8.MAX_LINE_LENGTH)
+
+    pep8style = pep8.StyleGuide(max_line_length=max_line_length)
     check = pep8style.input_file(fd.name)
     return check == 0
 
 
-def __jalopy(original, options):
+def __jalopy(original, options, use_nailgun=True):
+    # a temporary destination dir is needed with nailgun to prevent multiple jalopy instances from interfering
+    # with each other, a temporary directory
+    dest_dir = tempfile.mkdtemp('cvjalopy')
     jalopy_config = options.get('config')
     java_bin = options.get('java_bin', '/usr/bin/java')
+    ng_bin = options.get('ng_bin', '/usr/bin/ng-nailgun')
     classpath = options.get('classpath')
 
-    if not classpath:
-        raise ConfigurationError('Jalopy classpath not set')
+    if use_nailgun and os.path.isfile(ng_bin):
+        java_bin = ng_bin
+        # loglevel has to be WARN or otherwise we get exceptions when running multiple instances
+        jalopy = [java_bin, 'Jalopy', '--loglevel', 'WARN']
+    elif os.path.isfile(java_bin):
+        jalopy = [java_bin, '-classpath', classpath, 'Jalopy']
+        if not classpath:
+            raise ConfigurationError('Jalopy classpath not set')
+    else:
+        raise ConfigurationError('Jalopy java_bin option is invalid, %s does not exist' % java_bin)
 
     _env = {}
     _env.update(os.environ)
     _env['LANG'] = 'en_US.utf8'
     _env['LC_ALL'] = 'en_US.utf8'
-    with NamedTemporaryFile(suffix='.java', delete=False) as f:
-        f.write(original)
-        f.flush()
-        jalopy = [java_bin, '-classpath', classpath, 'Jalopy']
-        config = (['--convention', jalopy_config] if jalopy_config else [])
-        cmd = jalopy + config + [f.name]
-        j = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=_env)
-        stdout, stderr = j.communicate()
-        if stderr or '[ERROR]' in stdout:
-            raise ExecutionError('Failed to execute Jalopy: %s%s' % (stderr, stdout))
-        if '[WARN]' in stdout:
-            logging.info('Jalopy reports warnings: %s', stdout)
-        f.seek(0)
-        result = f.read()
+    try:
+        with NamedTemporaryFile(suffix='.java', delete=False) as f:
+            f.write(original)
+            f.flush()
+            destination = ['--flatdest', dest_dir]
+            config = (['--convention', jalopy_config] if jalopy_config else [])
+            cmd = jalopy + destination + config + ['--', f.name]
+            j = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=_env)
+            stdout, stderr = j.communicate()
+            if stderr or '[ERROR]' in stdout:
+                if stderr.strip().decode() == 'connect: Connection refused':
+                    # Fallback
+                    return __jalopy(original, options, use_nailgun=False)
+                raise ExecutionError('Failed to execute Jalopy: %s%s' % (stderr.decode(), stdout.decode()))
+            if '[WARN]' in stdout:
+                logging.info('Jalopy reports warnings: %s', stdout)
+            name = os.path.basename(f.name)
+            result = open(os.path.join(dest_dir, name)).read()
+    except:
+        result = ''
+    finally:
+        shutil.rmtree(dest_dir, True)
     return result
 
 
@@ -339,23 +397,28 @@ def _fix_pythontidy(src, dst):
     PythonTidy.tidy_up(src, dst)
 
 
-def _fix_pep8(src, dst, options):
+def _fix_pep8(src, dst, options={}):
     import autopep8
     if type(src) is file:
         source = src.read()
     else:
         source = src.getvalue()
 
-    class OptionsClass:
+    class OptionsClass(object):
 
-        select = options['select']
-        ignore = options['ignore']
-        pep8_passes = options['passes']
-        max_line_length = options['max_line_length']
+        '''Helper class for autopep8 options, just return None for unknown/new options'''
+
+        select = options.get('select')
+        ignore = options.get('ignore')
+        pep8_passes = options.get('passes')
+        max_line_length = options.get('max_line_length')
         verbose = False
         aggressive = True
 
-    fixed = autopep8.fix_string(source, options=OptionsClass)
+        def __getattr__(self, name):
+            return self.__dict__.get(name)
+
+    fixed = autopep8.fix_code(source, options=OptionsClass())
     dst.write(fixed)
 
 
@@ -405,7 +468,7 @@ def _validate_coffeelint(fd, options=None):
     """
 
     cfgfile = os.path.join(BASE_DIR, 'config/coffeelint.json')
-    po = subprocess.Popen('coffeelint --csv -s -f %s' % cfgfile, shell=True, stdin=subprocess.PIPE,
+    po = subprocess.Popen('coffeelint --reporter csv -s -f %s' % cfgfile, shell=True, stdin=subprocess.PIPE,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
     output, stderr = po.communicate(input=fd.read())
     valid = True
@@ -413,7 +476,7 @@ def _validate_coffeelint(fd, options=None):
         valid = False
         _detail(stderr)
     for row in output.split('\n'):
-        if row:
+        if row and row != 'path,lineNumber,lineNumberEnd,level,message':
             valid = False
             cols = row.split(',')
             if len(cols) > 3:
@@ -441,6 +504,45 @@ def _validate_puppet(fd):
         return valid
 
 
+@message('is not valid ruby')
+def _validate_ruby(fd):
+    p0 = subprocess.Popen(["ruby", "-c"], stdin=fd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    output, stderr = p0.communicate()
+    retcode = p0.poll()
+    if output.strip() != 'Syntax OK' or retcode != 0:
+        _detail("ruby parser exited with %d: %s" % (retcode, stderr))
+        return False
+    return True
+
+@message('is not rubocop formatted ruby code')
+def _validate_rubocop(fd):
+    p0 = subprocess.Popen(["rubocop", "--format", "emacs", fd.name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    output, stderr = p0.communicate()
+    retcode = p0.poll()
+    if retcode != 0:
+        _detail("rubocop exited with %d: \n%s" % (retcode, output))
+        return False
+    return True
+
+
+@message('is not valid ERB template')
+def _validate_erb(fd):
+    p1 = subprocess.Popen([
+        'erb',
+        '-P',
+        '-x',
+        '-T',
+        '-',
+    ], stdin=fd, stdout=subprocess.PIPE)
+    p2 = subprocess.Popen(['ruby', '-c'], stdin=p1.stdout, stdout=subprocess.PIPE)
+    p1.stdout.close()
+    output, stderr = p2.communicate()
+    retcode = p2.poll()
+    if output.strip() != 'Syntax OK' or retcode != 0:
+        return False
+    return True
+
+
 @message('has incomplete Maven POM description')
 def _validate_pomdesc(fd):
     """check Maven POM for title, description and organization"""
@@ -450,7 +552,7 @@ def _validate_pomdesc(fd):
     tree = ElementTree()
     try:
         elem = tree.parse(fd)
-    except Exception, e:
+    except Exception as e:
         _detail('%s: %s' % (e.__class__.__name__, e))
         return False
     # group = elem.findtext(NS + 'groupId')
@@ -477,25 +579,12 @@ def _validate_pomdesc(fd):
     return not VALIDATION_DETAILS
 
 
-@message('line comments ends with EOF')
-def _validate_sql_comment_last_line(fd, options={}):
-    sql_lines = fd.readlines()
-    last_line_is_a_comment = sql_lines[-1].startswith('--')
-    return not last_line_is_a_comment
-
-
-def _fix_sql_comment_last_line(src, dst, options={}):
-    original = src.read()
-    dst.write(original)
-    dst.write('\n')
-
-
 @message('SQL file ends without a semicolon')
 def _validate_sql_semi_colon(fd, options={}):
     import sqlparse
     sql = fd.read()
     sql_without_comments = sqlparse.format(sql, strip_comments=True).strip()
-    return sql_without_comments[-1] == ';'
+    return (sql_without_comments[-1] == ';' if sql_without_comments else True)
 
 
 def _fix_sql_semi_colon(src, dst, options={}):
@@ -508,14 +597,13 @@ def _fix_sql_semi_colon(src, dst, options={}):
 
 @message('doesn\'t pass Pyflakes validation')
 def _validate_pyflakes(fd, options={}):
-    from pyflakes import checker
-    tree = ast.parse(fd.read(), fd.name)
-    w = checker.Checker(tree, fd.name)
-    w.messages.sort(key=lambda x: x.lineno)
-    for message in w.messages:
+    proc = subprocess.Popen(['pyflakes', fd.name], stderr=subprocess.PIPE)
+    proc.wait()
+    errors = proc.stderr.read().decode().splitlines()
+    for message in errors:
         error = message.message % message.message_args
         _detail(error, line=message.lineno)
-    return len(w.messages) == 0
+    return proc.returncode == 0
 
 
 @message('contains syntax errors')
@@ -523,24 +611,33 @@ def _validate_database_dir(fname, options={}):
     if 'database/lounge' in fname or not fnmatch.fnmatch(fname, '*.sql'):
         return True
     pgsqlparser_bin = options.get('pgsql-parser-bin', '/opt/codevalidator/PgSqlParser')
+    if not os.path.isfile(pgsqlparser_bin):
+        raise ExecutionError('PostgreSQL parser binary not found, please set "pgsql-parser-bin" option')
 
     try:
-        return_code = subprocess.call([
-            pgsqlparser_bin,
-            '-q',
-            '-c',
-            '-i',
-            fname,
-        ])
+        with open(os.devnull, 'w') as devnull:
+            return_code = subprocess.call([
+                pgsqlparser_bin,
+                '-q',
+                '-c',
+                '-i',
+                fname,
+            ], stderr=devnull)
         return return_code == 0
     except:
         return False
 
 
 def _validate_sql_diff_dir(fname, options=None):
-    allowed_file_types = ['*.sql_diff', '*.py', '*.yml']
+    allowed_file_types = [
+        '*.sql_diff',
+        '*.py',
+        '*.yml',
+        '*.txt',
+        '*.md',
+    ]
     if not any(fnmatch.fnmatch(fname, each) for each in allowed_file_types):
-        return 'dbdiffs and migration scripts should use .sql_diff, .py or .yml extension'
+        return 'dbdiffs and migration scripts should use .sql_diff, .py, .yml, .md or .txt extension'
 
     dirs = get_dirs(fname)
     basedir = dirs[-2]
@@ -564,8 +661,8 @@ def _validate_sql_diff_sql(fname, options=None):
     sql = open(fname).read()
     has_set_role = re.search('[Ss][Ee][Tt] +[Rr][Oo][Ll][Ee] +[Tt][Oo] +zalando(_admin)?\s*', sql)
     has_set_project_schema_owner_role = \
-        re.search('''^ *[Ss][Ee][Ll][Ee][Cc][Tt] +[Zz][Zz]_[Uu][Tt][Ii][Ll][Ss].[Ss][Ee][Tt]_[Pp][Rr][Oo][Jj][Ee][Cc][Tt]_[Ss][Cc][Hh][Ee][Mm][Aa]_[Oo][Ww][Nn][Ee][Rr]_[Rr][Oo][Ll][Ee]\(\'\w+\'\);'''
-                  , sql)
+        re.search('''^ *select zz_utils\.set_project_schema_owner_role\('\w+'\);'''
+                  , sql, re.MULTILINE + re.IGNORECASE)
     if not (has_set_role or has_set_project_schema_owner_role):
         return 'set role to zalando; or SELECT zz_utils.set_project_schema_owner_role(); must be present in db diff'
 
@@ -638,7 +735,8 @@ def validate_file_dir_rules(fname):
                 res = func(fname, options)
             else:
                 res = func(fname)
-        except Exception, e:
+        except Exception as e:
+
             _error(fname, rule, func, 'ERROR validating {0}: {1}'.format(rule, e))
         else:
             if not res:
@@ -690,7 +788,7 @@ def validate_file_with_rules(fname, rules):
                     res = func(fd, options)
                 else:
                     res = func(fd)
-            except Exception, e:
+            except Exception as e:
                 _error(fname, rule, func, 'ERROR validating {0}: {1}'.format(rule, e))
             else:
                 if not res:
@@ -713,13 +811,25 @@ def validate_file(fname):
             validate_file_with_rules(fname, rules)
 
 
-def validate_directory(path):
+def validate_directory(path, exclude_patterns, include_patterns):
+    exclude_patterns = [os.path.join(path, pattern) for pattern in exclude_patterns or []]
+    include_patterns = [os.path.join(path, pattern) for pattern in include_patterns or []]
     for root, dirnames, filenames in os.walk(path):
         for exclude in CONFIG['exclude_dirs']:
             if exclude in dirnames:
                 dirnames.remove(exclude)
         for fname in filenames:
-            validate_file(os.path.join(root, fname))
+            fname = os.path.join(root, fname)
+            match_excluded = any(fnmatch.fnmatch(fname, pattern) for pattern in exclude_patterns)
+            match_included = any(fnmatch.fnmatch(fname, pattern) for pattern in include_patterns)
+
+            if exclude_patterns:
+                validate = not match_excluded or match_included
+            else:
+                validate = match_included or not include_patterns
+
+            if validate:
+                validate_file(fname)
 
 
 def fix_file(fname, rules):
@@ -743,7 +853,7 @@ def fix_file(fname, rules):
                     else:
                         func(src, dst)
                     was_fixed &= True
-                except Exception, e:
+                except Exception as e:
                     was_fixed = False
                     notify('{0}: ERROR fixing {1}: {2}'.format(fname, rule, e))
 
@@ -753,7 +863,7 @@ def fix_file(fname, rules):
     # b) some fix functions destroyed the code
     if was_fixed and len(fixed) > 0:
         with open_file_for_write(fname) as fd:
-            fd.write(fixed)
+            fd.write(fixed.encode())
         return True
     else:
         notify('{0}: ERROR fixing file. File remained unchanged'.format(fname))
@@ -779,7 +889,8 @@ def get_dirs(path):
 def main():
     parser = argparse.ArgumentParser(description='Validate source code files and optionally reformat them.')
     parser.add_argument('-r', '--recursive', action='store_true', help='process given directories recursively')
-    parser.add_argument('-c', '--config', help='use custom configuration file (default: ~/.codevalidatorrc)')
+    parser.add_argument('-c', '--config',
+                        help='use custom configuration file (default: ~/.codevalidatorrc or /etc/codevalidatorrc)')
     parser.add_argument('-f', '--fix', action='store_true', help='try to fix validation errors (by reformatting files)')
     parser.add_argument('-a', '--apply', metavar='RULE', action='append', help='apply the given rule(s)')
     parser.add_argument('-v', '--verbose', action='count', help='print more detailed error information (-vv for debug)')
@@ -787,14 +898,18 @@ def main():
     parser.add_argument('--filter', action='store_true',
                         help='special mode to read from STDIN and write to STDOUT, uses provided file name to find matching rules'
                         )
+    parser.add_argument('-e', '--exclude',  nargs='+', help='file patterns to exclude (only works with -r)')
+    parser.add_argument('-i', '--include',  nargs='+', help='file patterns to include (only works with -r)')
     parser.add_argument('files', metavar='FILES', nargs='+', help='list of source files to validate')
     args = parser.parse_args()
 
-    config_file = os.path.expanduser('~/.codevalidatorrc')
-    if os.path.isfile(config_file) and not args.config:
-        args.config = config_file
+    for path in DEFAULT_CONFIG_PATHS:
+        config_file = os.path.expanduser(path)
+        if os.path.isfile(config_file) and not args.config:
+            args.config = config_file
     if args.config:
-        CONFIG.update(json.load(open(args.config, 'rb')))
+        config = open(args.config, 'rb').read().decode()
+        CONFIG.update(json.loads(config))
     if args.verbose:
         CONFIG['verbose'] = args.verbose
         if args.verbose > 1:
@@ -828,7 +943,7 @@ def main():
 
         for f in args.files:
             if args.recursive and os.path.isdir(f):
-                validate_directory(f)
+                validate_directory(f, args.exclude, args.include)
             elif args.apply:
                 fix_file(f, args.apply)
             else:
